import 'dart:collection';

import 'package:flutter/foundation.dart';
import 'package:flutter/material.dart';
import 'package:flutter/rendering.dart';
import 'package:flutterx_application/flutterx_application.dart';
import 'package:flutterx_live_data/flutterx_live_data.dart';
import 'package:flutterx_preferences/flutterx_preferences.dart';
import 'package:flutterx_utils/flutterx_utils.dart';

part 'fragment.dart';
part 'fragment_manager.dart';

class FragmentContainer extends StatefulWidget {
  final AxisDirection axisDirection;
  final bool primary;
  final Storage? storage;
  final FragmentKey? initialFragmentKey;
  final Widget Function(Fragment fragment) fragmentWrapper;
  final Set<Fragment> fragments;

  @override
  FragmentContainerKey get key => super.key as FragmentContainerKey;

  const FragmentContainer({
    required FragmentContainerKey key,
    this.axisDirection = AxisDirection.right,
    this.primary = true,
    this.storage,
    this.initialFragmentKey,
    this.fragmentWrapper = removePrimaryScrollController,
    required this.fragments,
  })  : assert(fragments.length > 0, 'fragments cannot be empty'),
        super(key: key);

  @override
  State<FragmentContainer> createState() => FragmentContainerState();

  @override
  void debugFillProperties(DiagnosticPropertiesBuilder properties) {
    super.debugFillProperties(properties);
    properties
      ..add(EnumProperty<AxisDirection>('axisDirection', axisDirection))
      ..add(DiagnosticsProperty<bool>('primary', primary))
      ..add(DiagnosticsProperty<Storage?>('storage', storage))
      ..add(ObjectFlagProperty<Widget Function(Fragment)>.has('fragmentWrapper', fragmentWrapper))
      ..add(DiagnosticsProperty<FragmentKey?>('initialFragmentKey', initialFragmentKey));
  }

  static Widget removePrimaryScrollController(Fragment fragment) => PrimaryScrollController.none(child: fragment);

  static FragmentManager of(BuildContext context) => context.findAncestorStateOfType<FragmentContainerState>()!;

  static FragmentManager ofName(String name) => FragmentContainerKey(name).currentState!;
}

class FragmentContainerState extends State<FragmentContainer> implements FragmentManager {
  late PageController _controller;
  List<Fragment> _fragments = const [];
  Map<FragmentKey, int> _indexByKey = const {};
  ScrollDirection _direction = ScrollDirection.idle;

  @override
  FragmentContainer get container => widget;

  @override
  late FragmentStack currentFragment;

  @override
  LiveData<bool> get canGoBack => currentFragment.canUndo;

  @override
  LiveData<bool> get canGoForward => currentFragment.canRedo;

  @override
  void initState() {
    super.initState();
    _indexFragments();
    _initialize();
  }

  @override
  void didUpdateWidget(covariant FragmentContainer oldWidget) {
    super.didUpdateWidget(oldWidget);
    if (!setEquals(oldWidget.fragments, widget.fragments)) _indexFragments();
    if (oldWidget.key != widget.key || !identical(oldWidget.storage, widget.storage)) _initialize();
  }

  @override
  Widget build(BuildContext context) {
    if (!currentFragment.initialized) {
      currentFragment.value = _initialKey(context);
      _initController();
    }
    return Scrollable(
        key: ValueKey(widget.key),
        axisDirection: widget.axisDirection,
        controller: _controller,
        physics: const NeverScrollableScrollPhysics(),
        restorationId: widget.key.name,
        viewportBuilder: (context, position) =>
            Viewport(cacheExtent: 0, axisDirection: widget.axisDirection, offset: position, slivers: <Widget>[
              SliverFillViewport(
                  viewportFraction: _controller.viewportFraction,
                  delegate: SliverChildListDelegate(_fragments.map(widget.fragmentWrapper).toList(growable: false))),
            ]));
  }

  @override
  void navigateTo<T extends Fragment>({ScrollDirection direction = ScrollDirection.idle}) {
    for (final item in _fragments) {
      if (item is T) {
        _direction = direction;
        currentFragment.value = item.key;
        return;
      }
    }
    throw StateError('No fragment found with type $T');
  }

  @override
  void navigateToKey(FragmentKey key, {ScrollDirection direction = ScrollDirection.idle}) {
    _direction = direction;
    currentFragment.value = key;
  }

  @override
  bool back({ScrollDirection direction = ScrollDirection.reverse}) {
    if (currentFragment.canUndo.value) {
      _direction = direction;
      currentFragment.historyUndo();
      return true;
    }
    return false;
  }

  @override
  bool forward({ScrollDirection direction = ScrollDirection.forward}) {
    if (currentFragment.canRedo.value) {
      _direction = direction;
      currentFragment.historyRedo();
      return true;
    }
    return false;
  }

  bool _idle = false;

  void _indexFragments({List<FragmentKey>? order}) {
    _idle = order == null;
    _fragments = widget.fragments.toList(growable: false);
    if (order != null) _fragments.sort((a, b) => order.indexOf(a.key) - order.indexOf(b.key));
    _indexByKey = {for (var i = 0; i < _fragments.length; i++) _fragments[i].key: i};
  }

  FragmentKey? _lastKey;

  void _initialize() {
    currentFragment = FragmentStack._(key: '${widget.key.name}_active_fragment', storage: widget.storage)
      ..observeForever((key) {
        final last = _lastKey;
        if ((_direction != ScrollDirection.idle || !_idle) && last != null) {
          switch (_direction) {
            case ScrollDirection.idle:
              _indexFragments();
              break;
            case ScrollDirection.forward:
              _indexFragments(order: [last, key]);
              break;
            case ScrollDirection.reverse:
              _indexFragments(order: [key, last]);
              break;
          }
          final lastIndex = _indexByKey[last];
          if (lastIndex != null && lastIndex != _controller.page!.round())
            setState(() => _controller.jumpToPage(lastIndex));
          _direction = ScrollDirection.idle;
        }
        final index = _indexByKey[key];
        if (index != null) {
          _lastKey = key;
          if (widget.primary) AppPage.of(context).fragment = key;
          if (index != _controller.page!.round())
            _controller.animateToPage(index, duration: kThemeAnimationDuration, curve: Curves.ease);
        } else {
          currentFragment.value = _initialKey(context);
        }
      }, dispatcher: Dispatcher.postFrame);
    if (currentFragment.initialized) _initController();
  }

  FragmentKey _initialKey(BuildContext context) {
    if (widget.primary) {
      final key = AppPage.of(context).fragment;
      if (key != null && _indexByKey.containsKey(key)) return key;
    }
    final key = widget.initialFragmentKey;
    if (key != null && _indexByKey.containsKey(key)) return key;
    return _fragments.first.key;
  }

  void _initController() => _controller = PageController(initialPage: _indexByKey[currentFragment.value]!);
}

class FragmentStack extends MutableLiveData<FragmentKey> with HistoryMixin<FragmentKey> {
  final String _indexKey;
  final Storage? _storage;
  final Queue<String>? _queue;

  @override
  Queue<String> get queue => _queue ?? super.queue;

  FragmentStack._({required String key, required Storage? storage})
      : _indexKey = '${key}_index',
        _storage = storage,
        _queue = storage == null ? null : StorageLiveQueue(key: key, storage: storage),
        super.late() {
    if (storage != null) {
      final index = storage[_indexKey]?.let(int.parse) ?? -1;
      if (index != -1) {
        historyIndex = index;
        restoreState();
        updateValues();
      }
    }
  }

  @override
  void updateValues() {
    super.updateValues();
    final storage = _storage;
    if (storage != null) storage[_indexKey] = historyIndex.toString();
  }

  @override
  String serialize(FragmentKey data) => data.name;

  @override
  FragmentKey deserialize(String data) => FragmentKey(data);
}

class FragmentContainerKey extends GlobalKey<FragmentContainerState> {
  final String name;

  const FragmentContainerKey(this.name) : super.constructor();

  @override
  bool operator ==(Object other) => other is FragmentContainerKey && identical(other.name, name);

  @override
  int get hashCode => identityHashCode(name);

  @override
  String toString() => 'FragmentContainerKey($name)';
}
